import * as fs from 'node:fs';
import * as path from 'node:path';
import { STRUCTURE_FILES, EXTENSION, formatFile, getDataDir, extendErrorMessage, BUILDING_SYMBOLS_GUARD } from './common';
import { getStructure } from './map';
import { Struct } from './struct';

// Handle Errors
process.on('uncaughtException', function (e) {
    console.error('Error: ' + e.message);
    process.exit(1);
});

// Arguments
function nextArgument() {
    const out = process.argv.shift();
    if (!out) {
        throw new Error('Not Enough Arguments!');
    }
    return out;
}
nextArgument();
nextArgument();
// Output Files
function invalidFileType(file: string) {
    throw new Error(`Invalid File Type: ${file}`);
}
const sourceOutput = nextArgument();
fs.rmSync(sourceOutput, {force: true, recursive: true});
fs.mkdirSync(sourceOutput, {recursive: true});
const headerOutput = nextArgument();
if (!headerOutput.endsWith('.h')) {
    invalidFileType(headerOutput);
}
// Input Files
const extraHeaderFiles: string[] = [];
while (process.argv.length > 0) {
    const filePath = nextArgument();
    const file = path.parse(filePath);
    if (file.ext === EXTENSION) {
        // Get Class Name
        const fullName = file.base;
        const name = file.name;
        // Store
        if (STRUCTURE_FILES[name]) {
            throw new Error(`Multiple Definition Files Provided: ${fullName}`);
        }
        STRUCTURE_FILES[name] = filePath;
    } else if (file.ext === '.h') {
        // Extra Headers
        extraHeaderFiles.push(filePath);
    } else {
        // Invalid File Type
        invalidFileType(filePath);
    }
}

// Load And Sort Structures
function loadSymbols() {
    // Load
    const structureObjects = [];
    for (const name in STRUCTURE_FILES) {
        structureObjects.push(getStructure(name));
    }
    // Sort
    structureObjects.sort((a, b) => {
        if (a.name > b.name) {
            return 1;
        } else if (a.name < b.name) {
            return -1;
        } else {
            return 0;
        }
    });
    // Return
    return structureObjects;
}

// Sort Structures By Dependency
function dependencySort(structureObjects: Struct[]) {
    let loops = 0;
    const MAX_LOOPS = 1000;
    while (true) {
        if (loops > MAX_LOOPS) {
            throw new Error('Unable To Sort Dependencies');
        }
        loops++;
        let valid = true;
        // Loop Through Structures
        for (const structure of structureObjects) {
            // Loop Through Dependencies
            const currentIndex = structureObjects.indexOf(structure);
            for (const dependency of structure.getDependencies()) {
                // Compare Current And Dependency Index
                const obj = getStructure(dependency);
                const dependencyIndex = structureObjects.indexOf(obj);
                if (dependencyIndex > currentIndex) {
                    // Dependency Must Be Moved
                    structureObjects.splice(dependencyIndex, 1);
                    structureObjects.splice(currentIndex, 0, obj);
                    valid = false;
                    break;
                }
            }
            // Check If Any Changes Were Made
            if (!valid) {
                break;
            }
        }
        // Check If Any Changes Were Made
        if (valid) {
            break;
        }
    }
}

// Generate Part Of Header
function makeHeaderPart() {
    // Load Symbols
    const structureObjects = loadSymbols();

    // Sort Structures By Dependency
    dependencySort(structureObjects);

    // Generate Code
    let structures = '';
    for (const structure of structureObjects) {
        const name = structure.name;
        structures += `// ${name}\n`;
        try {
            structures += structure.generate();
        } catch (e) {
            throw new Error(extendErrorMessage(e, 'Generating Header: ' + name));
        }
        structures += '\n';
    }

    // Return
    return structures;
}

// Create Main Header
function makeMainHeader(output: string) {
    // Forward Declarations
    let forwardDeclarations = '';
    for (const name in STRUCTURE_FILES) {
        forwardDeclarations += `typedef struct ${name} ${name};\n`;
    }
    forwardDeclarations = forwardDeclarations.trim();
    // Extra Headers
    let extraHeaders = '';
    for (const file of extraHeaderFiles) {
        extraHeaders += fs.readFileSync(file, {encoding: 'utf8'});
    }
    extraHeaders = extraHeaders.trim();
    // Main
    const main = makeHeaderPart().trim();
    // Write
    const result = formatFile('out.h', {BUILDING_SYMBOLS_GUARD, forwardDeclarations, extraHeaders, main, data: getDataDir()});
    fs.writeFileSync(output, result);
}
makeMainHeader(headerOutput);

// Generate Compiled Code
function makeCompiledCode(outputDir: string) {
    // Load Symbols
    const structureObjects = loadSymbols();

    // Generate
    let first = true;
    for (const structure of structureObjects) {
        // Thunks
        let declarations = '';
        if (first) {
            first = false;
            declarations += '// Thunk Enabler\n';
            declarations += 'thunk_enabler_t thunk_enabler;\n\n';
        }

        // Structure
        const name = structure.name;
        declarations += `// ${name}\n`;
        try {
            declarations += structure.generateCode().trim();
        } catch (e) {
            throw new Error(extendErrorMessage(e, 'Generating Code: ' + name));
        }
        declarations += '\n';

        // Write
        const headerPath = fs.realpathSync(headerOutput);
        const main = declarations.trim();
        const result = formatFile('out.cpp', {BUILDING_SYMBOLS_GUARD, headerPath, main, data: getDataDir()});
        const output = path.join(outputDir, name + '.cpp');
        fs.writeFileSync(output, result);
    }
}
makeCompiledCode(sourceOutput);