import { COMMENT, extendErrorMessage, EXTENSION, parseTypeAndName, readDefinition, safeParseInt, syntaxError } from './common';
import { Method } from './method';
import { Property, StaticProperty } from './property';
import { Struct } from './struct';

// Error Handling
class ErrorOnLine extends Error {
    constructor(e: unknown, file: string, line: number) {
        // Get Message
        let message: string;
        if (e instanceof ErrorOnLine) {
            message = e.message;
        } else {
            message = extendErrorMessage(e, file + ':' + line.toString());
        }
        // Call Constructor
        super(message);
        Object.setPrototypeOf(this, ErrorOnLine.prototype);
    }
}

// Parse Property
export const OFFSET_SEPARATOR = ' = ';
export function parseProperty(args: string) {
    const [a, b] = args.split(OFFSET_SEPARATOR);
    if (!a || !b) {
        syntaxError('Invalid Piece Count');
    }
    const {type, name} = parseTypeAndName(a);
    const offset = safeParseInt(b);
    return {type, name, offset};
}

// Parse Method
export function parseMethod(args: string, self: string, insertSelfArg: boolean, isInherited: boolean) {
    const argsStart = args.indexOf('(');
    if (argsStart === -1) {
        syntaxError('Cannot Find Arguments');
    }
    const start = args.substring(0, argsStart).trim();
    const {type, name} = parseTypeAndName(start);
    const end = args.substring(argsStart).trim().split(OFFSET_SEPARATOR);
    if (!end[0] || !end[1]) {
        syntaxError('Invalid Piece Count');
    }
    const methodArgs = end[0].trim();
    if (!methodArgs.startsWith('(') || !methodArgs.endsWith(')')) {
        syntaxError('Invalid Method Arguments');
    }
    const address = safeParseInt(end[1]);
    return new Method(self, name, type, methodArgs, address, isInherited, !insertSelfArg);
}

// Load Structure
export function load(target: Struct, name: string, isExtended: boolean) {
    // Read File
    let data = readDefinition(name);

    // Strip Comments
    const lines = [];
    for (let line of data.split('\n')) {
        // Trim
        line = line.trim();
        // Remove Comments
        const index = line.indexOf(COMMENT);
        if (index !== -1) {
            line = line.substring(0, index);
        }
        // Store Line
        lines.push(line);
    }
    data = lines.join('\n');

    // Iterate Over Pieces
    let cursor = 0;
    for (let piece of data.split(';')) {
        // Find Start Of Command For Error Handling
        const startOfCommand = cursor + piece.search(/\S|$/);
        // Advance Cursor
        cursor += piece.length + 1;
        // Handle Errors
        try {
            // Trim
            piece = piece.trim();
            // Replace Newlines With Spaces
            piece = piece.replace(/\n/g, ' ');
            // Simplify
            piece = piece.replace(/\s+/g, ' ');
            // Skip Empty Piece
            if (piece.length === 0) {
                continue;
            }

            // Handle Commands
            let firstSpace = piece.indexOf(' ');
            if (firstSpace === -1) {
                firstSpace = piece.length;
            }
            const command = piece.substring(0, firstSpace);
            const args = piece.substring(firstSpace + 1);
            switch (command) {
                case 'extends': {
                    // Load Parent
                    if (!isExtended) {
                        target.setDirectParent(args);
                    }
                    load(target, args, true);
                    break;
                }
                case 'size': {
                    // Set Size
                    if (!isExtended) {
                        target.setSize(safeParseInt(args));
                    }
                    break;
                }
                case 'vtable-size': {
                    // Set VTable Size
                    if (!isExtended) {
                        target.getVTable().setSize(safeParseInt(args));
                    }
                    break;
                }
                case 'vtable': {
                    // Set VTable Address
                    const vtable = target.getVTable();
                    if (!isExtended && args.length > 0) {
                        vtable.setAddress(safeParseInt(args));
                    }
                    break;
                }
                case 'property': {
                    // Add Property
                    const info = parseProperty(args);
                    target.addProperty(new Property(info.offset, info.type, info.name, target.name));
                    break;
                }
                case 'static-property': {
                    // Add Static Property
                    if (!isExtended) {
                        const info = parseProperty(args);
                        target.addStaticProperty(new StaticProperty(info.offset, info.type, info.name, target.name));
                    }
                    break;
                }
                case 'method': {
                    // Add Method
                    const method = parseMethod(args, name, true, isExtended);
                    target.addMethod(method, false, false);
                    break;
                }
                case 'virtual-method': {
                    // Add Virtual Method
                    const method = parseMethod(args, target.name, true, isExtended);
                    target.addMethod(method, true, false);
                    break;
                }
                case 'static-method': {
                    // Add Static Method
                    if (!isExtended) {
                        const method = parseMethod(args, target.name, false, false);
                        target.addMethod(method, false, false);
                    }
                    break;
                }
                case 'constructor': {
                    // Constructor
                    if (!isExtended) {
                        let data = target.name + ' *constructor';
                        if (args.startsWith('(')) {
                            // No Custom Name
                        } else {
                            // Use Custom Name
                            data += '_';
                        }
                        data += args;
                        const method = parseMethod(data, target.name, true, false);
                        target.addMethod(method, false, true);
                    }
                    break;
                }
                case 'vtable-destructor-offset': {
                    // Set VTable Destructor Offset
                    target.getVTable().setDestructorOffset(safeParseInt(args));
                    break;
                }
                case 'mark-as-simple': {
                    // Mark As Simple
                    if (isExtended) {
                        throw new Error('Cannot Extend Simple Structure');
                    }
                    target.markAsSimple();
                    break;
                }
                case 'generate-custom-wrapper': {
                    // Custom Wrapper
                    if (!isExtended) {
                        target.enableCustomWrapper();
                    }
                    break;
                }
                default: {
                    throw new Error(`Invalid Command: ${command}`);
                }
            }
        } catch (e) {
            // Find Line Number
            let lineNumber = 1;
            for (let i = 0; i <= startOfCommand; i++) {
                if (data.charAt(i) === '\n') {
                    lineNumber++;
                }
            }
            // Rethrow Error With Line Number
            throw new ErrorOnLine(e, `${name}${EXTENSION}`, lineNumber);
        }
    }
}